# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from freediscovery.server import fd_app

app = fd_app('/freediscovery_shared/')
